#ifndef SHRINK_MATRIX_VIEW_HPP
#define SHRINK_MATRIX_VIEW_HPP

#include <utils/PugsAssert.hpp>
#include <utils/PugsMacros.hpp>

#include <cstddef>
#include <iostream>
#include <utils/NaNHelper.hpp>

template <typename MatrixType>
class ShrinkMatrixView
{
 public:
  using index_type = typename MatrixType::index_type;
  using data_type  = typename MatrixType::data_type;

 private:
  MatrixType& m_matrix;
  const size_t m_nb_rows;

 public:
  friend std::ostream&
  operator<<(std::ostream& os, const ShrinkMatrixView& A)
  {
    for (size_t i = 0; i < A.numberOfRows(); ++i) {
      os << i << '|';
      for (size_t j = 0; j < A.numberOfColumns(); ++j) {
        os << ' ' << j << ':' << NaNHelper(A(i, j));
      }
      os << '\n';
    }
    return os;
  }

  PUGS_INLINE size_t
  numberOfRows() const noexcept
  {
    return m_nb_rows;
  }

  PUGS_INLINE size_t
  numberOfColumns() const noexcept
  {
    return m_matrix.numberOfColumns();
  }

  PUGS_INLINE
  data_type&
  operator()(index_type i, index_type j) const noexcept(NO_ASSERT)
  {
    Assert(i < m_nb_rows and j < m_matrix.numberOfColumns(), "cannot access element: invalid indices");
    return m_matrix(i, j);
  }

  ShrinkMatrixView(MatrixType& matrix, size_t nb_rows) noexcept(NO_ASSERT) : m_matrix{matrix}, m_nb_rows{nb_rows}
  {
    Assert(m_nb_rows <= matrix.numberOfRows(), "shrink number of rows must be smaller than original matrix's");
  }

  ShrinkMatrixView(const ShrinkMatrixView&) = delete;
  ShrinkMatrixView(ShrinkMatrixView&&)      = delete;

  ~ShrinkMatrixView() noexcept = default;
};

#endif   // SHRINK_MATRIX_VIEW_HPP
