#ifndef ACOUSTIC_COMPOSITE_SOLVER_HPP
#define ACOUSTIC_COMPOSITE_SOLVER_HPP

#include <mesh/MeshTraits.hpp>

#include <memory>
#include <tuple>
#include <vector>

class DiscreteFunctionVariant;
class IBoundaryConditionDescriptor;
class MeshVariant;
class ItemValueVariant;
class SubItemValuePerItemVariant;

double composite_acoustic_dt(const std::shared_ptr<const DiscreteFunctionVariant>& c);

class AcousticCompositeSolverHandler
{
 public:
  enum class SolverType
  {
    GlaceComposite,
    EucclhydComposite
  };
  
 private:
  struct IAcousticCompositeSolver
  {

    virtual std::tuple<const std::shared_ptr<const ItemValueVariant>,
		       const std::shared_ptr<const SubItemValuePerItemVariant>,
		       const std::shared_ptr<const ItemValueVariant>,
		       const std::shared_ptr<const SubItemValuePerItemVariant>,
		       const std::shared_ptr<const ItemValueVariant>,
		       const std::shared_ptr<const SubItemValuePerItemVariant>
		       >
    
    compute_fluxes(
      const SolverType& solver_type,
      const std::shared_ptr<const DiscreteFunctionVariant>& rho,
      const std::shared_ptr<const DiscreteFunctionVariant>& c,
      const std::shared_ptr<const DiscreteFunctionVariant>& u,
      const std::shared_ptr<const DiscreteFunctionVariant>& p,
      const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>& bc_descriptor_list) const = 0;


    virtual std::tuple<std::shared_ptr<const MeshVariant>,
                       std::shared_ptr<const DiscreteFunctionVariant>,
                       std::shared_ptr<const DiscreteFunctionVariant>,
                       std::shared_ptr<const DiscreteFunctionVariant>>
    apply_fluxes(const double& dt,
                 const std::shared_ptr<const DiscreteFunctionVariant>& rho,
                 const std::shared_ptr<const DiscreteFunctionVariant>& u,
                 const std::shared_ptr<const DiscreteFunctionVariant>& E,
                 const std::shared_ptr<const ItemValueVariant>& ur,
		 const std::shared_ptr<const SubItemValuePerItemVariant>& Fjr,
	         const std::shared_ptr<const ItemValueVariant>& ue,
		 const std::shared_ptr<const SubItemValuePerItemVariant>& Fje,
		 const std::shared_ptr<const ItemValueVariant>& uf,
		 const std::shared_ptr<const SubItemValuePerItemVariant>& Fjf) const = 0;
    
    
    virtual std::tuple<std::shared_ptr<const MeshVariant>,
                       std::shared_ptr<const DiscreteFunctionVariant>,
                       std::shared_ptr<const DiscreteFunctionVariant>,
                       std::shared_ptr<const DiscreteFunctionVariant>>
    apply(const SolverType& solver_type,
          const double& dt,
          const std::shared_ptr<const DiscreteFunctionVariant>& rho,
          const std::shared_ptr<const DiscreteFunctionVariant>& u,
          const std::shared_ptr<const DiscreteFunctionVariant>& E,
          const std::shared_ptr<const DiscreteFunctionVariant>& c,
          const std::shared_ptr<const DiscreteFunctionVariant>& p,
          const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>& bc_descriptor_list) const = 0;

    IAcousticCompositeSolver()                  = default;
    IAcousticCompositeSolver(IAcousticCompositeSolver&&) = default;
    IAcousticCompositeSolver& operator=(IAcousticCompositeSolver&&) = default;

    virtual ~IAcousticCompositeSolver() = default;
  };

  template <MeshConcept MeshType>
  class AcousticCompositeSolver;

  std::unique_ptr<IAcousticCompositeSolver> m_acoustic_composite_solver;

 public:
  const IAcousticCompositeSolver&
  solver() const
  {
    return *m_acoustic_composite_solver;
  }

  AcousticCompositeSolverHandler(const std::shared_ptr<const MeshVariant>& mesh_v);
};

#endif   // ACOUSTIC_COMPOSITE_SOLVER_HPP
